# -*- coding: utf-8 -*-
# Generated by Django 1.9.7 on 2016-06-27 21:00
from __future__ import unicode_literals

import django.db.models.deletion
from django.db import migrations, models


class CreateConcreteModelFromProxy(migrations.CreateModel):
    def state_forwards(self, app_label, state):
        state.remove_model(app_label, self.name_lower)
        super(CreateConcreteModelFromProxy, self).state_forwards(app_label, state)


class Migration(migrations.Migration):

    dependencies = [
        ('related', '0002_update_field_defs_app_label'),
    ]

    operations = [
        CreateConcreteModelFromProxy(
            name='OneToOneFieldDefinition',
            fields=[
                ('foreignkeydefinition_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='related.ForeignKeyDefinition')),
                ('parent_link', models.BooleanField(default=False, verbose_name='parent link')),
            ],
            bases=('related.foreignkeydefinition',),
        ),
    ]



from __future__ import unicode_literals

import logging

from django.utils.version import get_version

VERSION = (0, 3, 0, 'alpha', 3)

__version__ = get_version(VERSION)

logger = logging.getLogger('mutant')

default_app_config = 'mutant.apps.MutantConfig'



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('mutant', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='GeometryFieldDefinition',
            fields=[
                ('fielddefinition_ptr', models.OneToOneField(
                    to='mutant.FieldDefinition', on_delete=models.CASCADE, parent_link=True, auto_created=True,
                    primary_key=True, serialize=False
                )),
                ('srid', models.IntegerField(default=4326, help_text='Spatial Reference System Identity', verbose_name='SRID')),
                ('spatial_index', models.BooleanField(default=True, help_text='Creates a spatial index for the given geometry field.', verbose_name='spatial index')),
                ('dim', models.PositiveSmallIntegerField(default=2, help_text='Coordinate dimension.', verbose_name='coordinate dimension', choices=[(2, 'Two-dimensional'), (3, 'Three-dimensional')])),
                ('geography', models.BooleanField(default=False, help_text='Creates a database column of type geography, rather than geometry.', verbose_name='geography')),
            ],
            options={
                'db_table': 'mutant_geometryfielddefinition',
            },
            bases=('mutant.fielddefinition',),
        ),
        migrations.CreateModel(
            name='GeometryCollectionFieldDefinition',
            fields=[
            ],
            options={
                'proxy': True,
            },
            bases=('geo.geometryfielddefinition',),
        ),
        migrations.CreateModel(
            name='LineStringFieldDefinition',
            fields=[
            ],
            options={
                'proxy': True,
            },
            bases=('geo.geometryfielddefinition',),
        ),
        migrations.CreateModel(
            name='MultiLineStringFieldDefinition',
            fields=[
            ],
            options={
                'proxy': True,
            },
            bases=('geo.geometryfielddefinition',),
        ),
        migrations.CreateModel(
            name='MultiPointFieldDefinition',
            fields=[
            ],
            options={
                'proxy': True,
            },
            bases=('geo.geometryfielddefinition',),
        ),
        migrations.CreateModel(
            name='MultiPolygonFieldDefinition',
            fields=[
            ],
            options={
                'proxy': True,
            },
            bases=('geo.geometryfielddefinition',),
        ),
        migrations.CreateModel(
            name='PointFieldDefinition',
            fields=[
            ],
            options={
                'proxy': True,
            },
            bases=('geo.geometryfielddefinition',),
        ),
        migrations.CreateModel(
            name='PolygonFieldDefinition',
            fields=[
            ],
            options={
                'proxy': True,
            },
            bases=('geo.geometryfielddefinition',),
        ),
    ]



from __future__ import unicode_literals

from django.db.models import Q
from django.db.models.fields.related import RelatedField
from django.utils.six import string_types

from ....compat import get_remote_field_model
from ....db.models import MutableModel
from ....models import ModelDefinition
from ....utils import clear_opts_related_cache


def mutable_model_prepared(signal, sender, definition, existing_model_class,
                           **kwargs):
    """
    Make sure all related model class are created and marked as dependency
    when a mutable model class is prepared
    """
    referenced_models = set()
    # Collect all model class the obsolete model class was referring to
    if existing_model_class:
        for field in existing_model_class._meta.local_fields:
            if isinstance(field, RelatedField):
                remote_field_model = get_remote_field_model(field)
                if not isinstance(remote_field_model, string_types):
                    referenced_models.add(remote_field_model)
    # Add sender as a dependency of all mutable models it refers to
    for field in sender._meta.local_fields:
        if isinstance(field, RelatedField):
            remote_field_model = get_remote_field_model(field)
            if not isinstance(remote_field_model, string_types):
                referenced_models.add(remote_field_model)
                if (issubclass(remote_field_model, MutableModel) and
                        remote_field_model._definition != sender._definition):
                    remote_field_model._dependencies.add(sender._definition)
    # Mark all model referring to this one as dependencies
    related_model_defs = ModelDefinition.objects.filter(
        Q(fielddefinitions__foreignkeydefinition__to=definition) |
        Q(fielddefinitions__manytomanyfielddefinition__to=definition)
    ).distinct()
    for model_def in related_model_defs:
        if model_def != definition:
            # Generate model class from definition and add it as a dependency
            sender._dependencies.add(model_def.model_class()._definition)
    # Clear the referenced models opts related cache
    for model_class in referenced_models:
        clear_opts_related_cache(model_class)



from __future__ import unicode_literals

import json
import logging
import math
import time
from threading import Thread

import redis
from django.utils.encoding import force_str

logger = logging.getLogger(__name__)


class BaseEngine(Thread):
    def __init__(self, initialize, callback):
        self.initialize = initialize
        self.callback = callback
        self.ready = False
        super(BaseEngine, self).__init__(name="%s.%s" % (self.__module__, self.__class__.__name__))

    def _initialize(self):
        self.initialize()
        self.ready = True

    def stop(self, timeout=None):
        self.join(timeout)

    def publish(self, *args):
        raise NotImplementedError


class Redis(BaseEngine):
    def __init__(self, initialize, callback, channel='mutant-state',
                 socket_timeout=60*10, socket_keepalive=True, **options):
        super(Redis, self).__init__(initialize, callback)
        self.redis = redis.StrictRedis(
            socket_timeout=socket_timeout,
            socket_keepalive=socket_keepalive,
            retry_on_timeout=False,
            **options
        )
        self.socket_timeout = socket_timeout
        self.channel = channel
        self.pubsub = self.redis.pubsub()
        self.pubsub.subscribe(self.channel)

    def _run(self):
        self._initialize()
        for event in self.pubsub.listen():
            if event['type'] == 'message':
                args = json.loads(force_str(event['data']))
                self.callback(*args)

    def _reconnect(self):
        connection = self.pubsub.connection
        if connection:
            connection.disconnect()
        while True:
            logger.info('Attempting to reconnect.')
            try:
                connection.connect()
            except redis.ConnectionError:
                logger.exception('Failed to reconnect, will re-attempt in 5 seconds.')
                time.sleep(5)
            else:
                logger.info('Successfully reconnected.')
                break

    def run(self):
        start = time.time()
        try:
            self._run()
        except redis.TimeoutError:
            if math.ceil(time.time() - start) < self.socket_timeout:
                logger.warning('Unexpected connection timeout.')
            self._reconnect()
            self.run()
        except redis.ConnectionError:
            logger.warning('Connection error.')
            self._reconnect()
            self.run()

    def publish(self, *args):
        message = json.dumps(args)
        self.redis.publish(self.channel, message)

    def join(self, timeout=None):
        self.pubsub.unsubscribe(self.channel)
        super(Redis, self).join(timeout=None)



from __future__ import unicode_literals

from django.contrib.contenttypes.models import ContentType
from django.db import connections, router, transaction
from django.db.utils import IntegrityError
from django.test.testcases import TestCase

from ..models.model import ModelDefinition
from ..utils import remove_from_app_cache


def connections_can_rollback_ddl():
    """
    Returns True if all implied connections have DDL transactions support.
    """
    return all(connection.features.can_rollback_ddl for connection in connections.all())


class DDLTestCase(TestCase):
    """
    A class that behaves like `TestCase` if all connections support DDL
    transactions or like `TransactionTestCase` if it's not the case.
    """
    manual_transaction = False

    @classmethod
    def _use_transactions(cls):
        return not cls.manual_transaction and connections_can_rollback_ddl()

    @classmethod
    def setUpClass(cls):
        if cls._use_transactions():
            return super(DDLTestCase, cls).setUpClass()
        else:
            return super(TestCase, cls).setUpClass()

    @classmethod
    def tearDownClass(cls):
        if cls._use_transactions():
            return super(DDLTestCase, cls).tearDownClass()
        else:
            return super(TestCase, cls).tearDownClass()

    def _should_reload_connections(self):
        if self._use_transactions():
            return super(DDLTestCase, self)._should_reload_connections()
        else:
            return super(TestCase, self)._should_reload_connections()

    def _fixture_setup(self):
        if self._use_transactions():
            return super(DDLTestCase, self)._fixture_setup()
        else:
            self.setUpTestData()
            return super(TestCase, self)._fixture_setup()

    def _fixture_teardown(self):
        if self._use_transactions():
            return super(DDLTestCase, self)._fixture_teardown()
        else:
            return super(TestCase, self)._fixture_teardown()


class ModelDefinitionDDLTestCase(DDLTestCase):
    def tearDown(self):
        if not self._use_transactions():
            # Remove all the extra tables since `TransactionTestCase` only
            # truncate data on teardown.
            ModelDefinition.objects.all().delete()
        else:
            for model_def in ModelDefinition.objects.all():
                remove_from_app_cache(model_def.model_class())
        ContentType.objects.clear_cache()


class FieldDefinitionTestMixin(object):
    field_definition_init_kwargs = {}
    field_values = ()

    @classmethod
    def setUpTestData(cls):
        super(FieldDefinitionTestMixin, cls).setUpTestData()
        with cls.assertChecksumChange():
            cls.field_pk = cls.field_definition_cls._default_manager.create(
                model_def_id=cls.model_def_pk, name='field',
                **cls.field_definition_init_kwargs
            ).pk

    def setUp(self):
        super(FieldDefinitionTestMixin, self).setUp()
        self.field = self.field_definition_cls._default_manager.get(pk=self.field_pk)

    def get_field_value(self, instance, name='field'):
        return getattr(instance, name)

    def prepare_default_value(self, value):
        return value

    def test_field_default(self):
        default = self.prepare_default_value(self.field_values[0])
        field = self.field
        # Default value should be valid
        with self.assertChecksumChange():
            field.default = default
            field.full_clean()
            field.save()
        # Default value should be assigned correctly
        model_class = self.model_def.model_class()
        instance = model_class.objects.create()
        field_value = self.get_field_value(instance)
        created_default = self.prepare_default_value(field_value)
        self.assertEqual(created_default, default)

    def test_create_with_default(self):
        """Makes sure a field definition manager is attached to the model and
        `create_with_default` works correctly."""
        model_class = self.model_def.model_class()
        field_value = self.field_values[0]
        instance = model_class.objects.create(field=field_value)
        # Add the field with a default.
        create_default = self.prepare_default_value(field_value)
        options = dict(**self.field_definition_init_kwargs)
        options['default'] = create_default
        with self.assertChecksumChange():
            self.field_definition_cls._default_manager.create_with_default(
                model_def=self.model_def, name='field_created_with_default',
                **options
            )
        created_value = self.prepare_default_value(
            model_class.objects.get(pk=instance.pk).field_created_with_default
        )
        self.assertEqual(created_value, create_default)

    def test_model_save(self):
        first_value, second_value = self.field_values
        # Assigning a value should work
        model_class = self.model_def.model_class()
        instance = model_class.objects.create(field=first_value)
        self.assertEqual(self.get_field_value(instance), first_value)
        # Assigning a new one should also work
        instance.field = second_value
        instance.save()
        instance = model_class.objects.get()
        self.assertEqual(self.get_field_value(instance), second_value)

    def test_field_renaming(self):
        value = self.field_values[0]
        model_class = self.model_def.model_class()
        # Renaming a field should update its column name
        model_class.objects.create(field=value)
        opts = model_class._meta
        original_column_name = opts.get_field('field').get_attname_column()[1]
        with self.assertChecksumChange():
            self.field.name = 'renamed_field'
            self.field.save()
        opts = model_class._meta
        new_column_name = opts.get_field('renamed_field').get_attname_column()[1]
        self.assertModelTablesColumnDoesntExists(
            model_class, original_column_name
        )
        self.assertModelTablesColumnExists(
            model_class, new_column_name
        )
        # Old data should be accessible by the new field name
        instance = model_class.objects.get()
        self.assertEqual(self.get_field_value(instance, 'renamed_field'), value)
        # The old field shouldn't be accessible anymore
        msg = "'field' is an invalid keyword argument for this function"
        self.assertRaisesMessage(TypeError, msg, model_class, field=value)
        # It should be possible to create objects using the new field name
        model_class.objects.create(renamed_field=value)

    def test_field_deletion(self):
        value = self.field_values[0]
        model_class = self.model_def.model_class()
        model_class.objects.create(field=value)
        # Deleting a field should delete the associated column
        opts = model_class._meta
        field_column_name = opts.get_field('field').get_attname_column()[1]
        with self.assertChecksumChange():
            self.field.delete()
        self.assertModelTablesColumnDoesntExists(model_class, field_column_name)
        # The deleted field shouldn't be accessible anymore
        msg = "'field' is an invalid keyword argument for this function"
        self.assertRaisesMessage(TypeError, msg, model_class, field=value)

    def test_field_unique(self):
        value = self.field_values[0]
        model_class = self.model_def.model_class()
        with self.assertChecksumChange():
            self.field.unique = True
            self.field.save()
        model_class.objects.create(field=value)
        using = router.db_for_write(model_class)
        try:
            with transaction.atomic(using, savepoint=True):
                model_class.objects.create(field=value)
        except IntegrityError:
            pass
        else:
            self.fail("One shouldn't be able to save duplicate entries in a unique field")

    def test_field_cloning(self):
        with self.assertChecksumChange():
            clone = self.field.clone()
            clone.name = 'field_clone'
            clone.model_def = self.model_def
            clone.save(force_insert=True)

    def test_field_definition_category(self):
        self.assertEqual(
            self.field_definition_cls.get_field_category(),
            self.field_definition_category
        )



from __future__ import unicode_literals

from django.core.exceptions import ValidationError
from django.db import models
from django.db.migrations.state import ModelState, StateApps
from django.utils.six import string_types
from django.utils.translation import ugettext_lazy as _

from .. import logger
from ..compat import get_remote_field_model
from ..state import handler as state_handler


class MutableModel(models.Model):
    """Abstract class used to identify models that we're created by a
    definition."""

    class Meta:
        abstract = True

    @classmethod
    def definition(cls):
        definition_cls, definition_pk = cls._definition
        return definition_cls.objects.get(pk=definition_pk)

    @classmethod
    def checksum(cls):
        return cls._checksum

    @classmethod
    def is_obsolete(cls):
        return (
            cls._is_obsolete or
            cls._checksum != state_handler.get_checksum(cls._definition[1])
        )

    @classmethod
    def get_model_state(cls, **kwargs):
        return ModelState.from_model(cls, **kwargs)

    @classmethod
    def get_related_model_states(cls, model_state):
        model_states = {}
        for _name, field in model_state.fields:
            related_model_reference = get_remote_field_model(field)
            if related_model_reference:
                related_model = cls._meta.apps.get_model(related_model_reference)
                if issubclass(related_model, MutableModel):
                    related_model_state = related_model.get_model_state(exclude_rels=True)
                else:
                    related_model_state = ModelState.from_model(related_model, exclude_rels=True)
                model_states[related_model_state.app_label, related_model_state.name] = related_model_state
                for base in related_model_state.bases:
                    if isinstance(base, string_types):
                        base_model = cls._meta.apps.get_model(base)
                        if issubclass(base_model, MutableModel):
                            base_model_state = base_model.get_model_state(exclude_rels=True)
                        else:
                            base_model_state = ModelState.from_model(base_model, exclude_rels=True)
                        model_states[base_model_state.app_label, base_model_state.name] = base_model_state
        return list(model_states.values())

    @classmethod
    def render_state(cls):
        state = cls.get_model_state()
        related_states = cls.get_related_model_states(state)
        apps = StateApps([], {})
        apps.render_multiple(related_states + [state])
        return apps.all_models[state.app_label][state.name.lower()]

    @classmethod
    def mark_as_obsolete(cls, origin=None):
        cls._is_obsolete = True
        logger.debug(
            "Marking model %s and it dependencies (%s) as obsolete.",
            cls, cls._dependencies
        )
        if origin is None:
            origin = cls._definition
        for definition_cls, definition_pk in cls._dependencies:
            if (definition_cls, definition_pk) == origin:
                continue
            try:
                definition = definition_cls.objects.get(pk=definition_pk)
            except definition_cls.DoesNotExist:
                pass
            else:
                definition.model_class().mark_as_obsolete(origin)

    def clean(self):
        if self.is_obsolete():
            raise ValidationError('Obsolete definition')
        return super(MutableModel, self).clean()

    def save(self, *args, **kwargs):
        if self.is_obsolete():
            msg = _('Cannot save an obsolete model')
            raise ValidationError(msg)
        return super(MutableModel, self).save(*args, **kwargs)

    def delete(self, *args, **kwargs):
        if self.is_obsolete():
            msg = _('Cannot delete an obsolete model')
            raise ValidationError(msg)
        return super(MutableModel, self).delete(*args, **kwargs)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('mutant', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='DateFieldDefinition',
            fields=[
                ('fielddefinition_ptr', models.OneToOneField(
                    to='mutant.FieldDefinition', on_delete=models.CASCADE, parent_link=True, auto_created=True,
                    primary_key=True, serialize=False
                )),
                ('auto_now', models.BooleanField(default=False, help_text='Automatically set the field to now every time the object is saved.', verbose_name='auto now')),
                ('auto_now_add', models.BooleanField(default=False, help_text='Automatically set the field to now when the object is first created.', verbose_name='auto now add')),
            ],
            options={
                'db_table': 'mutant_datefielddefinition',
            },
            bases=('mutant.fielddefinition',),
        ),
        migrations.CreateModel(
            name='DateTimeFieldDefinition',
            fields=[
            ],
            options={
                'proxy': True,
            },
            bases=('temporal.datefielddefinition',),
        ),
        migrations.CreateModel(
            name='TimeFieldDefinition',
            fields=[
            ],
            options={
                'proxy': True,
            },
            bases=('temporal.datefielddefinition',),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import picklefield.fields
from django.db import migrations, models

import mutant.db.fields.python


class Migration(migrations.Migration):

    dependencies = [
        ('mutant', '0001_initial'),
        ('contenttypes', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='ForeignKeyDefinition',
            fields=[
                ('fielddefinition_ptr', models.OneToOneField(
                    to='mutant.FieldDefinition', on_delete=models.CASCADE, parent_link=True, auto_created=True,
                    primary_key=True, serialize=False
                )),
                ('related_name', mutant.db.fields.python.PythonIdentifierField(help_text='The name to use for the relation from the related object back to this one.', max_length=255, null=True, verbose_name='related name', blank=True)),
                ('to_field', mutant.db.fields.python.PythonIdentifierField(help_text='The field on the related object that the relation is to.', max_length=255, null=True, verbose_name='to field', blank=True)),
                ('one_to_one', models.BooleanField(default=False, editable=False)),
                ('on_delete', models.CharField(default='CASCADE', choices=[('CASCADE', 'CASCADE'), ('PROTECT', 'PROTECT'), ('SET_NULL', 'SET_NULL'), ('SET_DEFAULT', 'SET_DEFAULT'), ('SET_VALUE', 'SET(VALUE)'), ('DO_NOTHING', 'DO_NOTHING')], max_length=11, blank=True, help_text='Behavior when an object referenced by this field is deleted', null=True, verbose_name='on delete')),
                ('on_delete_set_value', picklefield.fields.PickledObjectField(verbose_name='on delete set value', null=True, editable=False)),
                ('to', models.ForeignKey(
                    to='contenttypes.ContentType', on_delete=models.CASCADE, related_name='+', verbose_name='to'
                )),
            ],
            options={
                'db_table': 'mutant_foreignkeydefinition',
            },
            bases=('mutant.fielddefinition',),
        ),
        migrations.CreateModel(
            name='ManyToManyFieldDefinition',
            fields=[
                ('fielddefinition_ptr', models.OneToOneField(
                    to='mutant.FieldDefinition', on_delete=models.CASCADE, parent_link=True, auto_created=True,
                    primary_key=True, serialize=False
                )),
                ('related_name', mutant.db.fields.python.PythonIdentifierField(help_text='The name to use for the relation from the related object back to this one.', max_length=255, null=True, verbose_name='related name', blank=True)),
                ('symmetrical', models.NullBooleanField(verbose_name='symmetrical')),
                ('db_table', models.SlugField(help_text='The name of the table to create for storing the many-to-many data', max_length=30, null=True, blank=True)),
                ('through', models.ForeignKey(
                    to='contenttypes.ContentType', on_delete=models.CASCADE, related_name='+', blank=True, help_text='Intermediary model', null=True)),
                ('to', models.ForeignKey(
                    to='contenttypes.ContentType', on_delete=models.CASCADE, related_name='+', verbose_name='to'
                )),
            ],
            options={
                'db_table': 'mutant_manytomanyfielddefinition',
            },
            bases=('mutant.fielddefinition',),
        ),
        migrations.CreateModel(
            name='OneToOneFieldDefinition',
            fields=[
            ],
            options={
                'proxy': True,
            },
            bases=('related.foreignkeydefinition',),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import picklefield.fields
import polymodels.fields
from django.db import migrations, models

import mutant.db.deletion
import mutant.db.fields.generic
import mutant.db.fields.python
import mutant.db.fields.translation
import mutant.models.field


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='BaseDefinition',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('order', models.PositiveIntegerField(editable=False)),
                ('base', picklefield.fields.PickledObjectField(verbose_name='base', editable=False)),
            ],
            options={
                'ordering': ['order'],
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='FieldDefinition',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', mutant.db.fields.python.PythonIdentifierField(max_length=255, verbose_name='name')),
                ('verbose_name', mutant.db.fields.translation.LazilyTranslatedField(null=True, verbose_name='verbose name', blank=True)),
                ('help_text', mutant.db.fields.translation.LazilyTranslatedField(null=True, verbose_name='help text', blank=True)),
                ('null', models.BooleanField(default=False, verbose_name='null')),
                ('blank', models.BooleanField(default=False, verbose_name='blank')),
                ('db_column', models.SlugField(max_length=30, null=True, verbose_name='db column', blank=True)),
                ('db_index', models.BooleanField(default=False, verbose_name='db index')),
                ('editable', models.BooleanField(default=True, verbose_name='editable')),
                ('default', picklefield.fields.PickledObjectField(default=mutant.models.field.NOT_PROVIDED, verbose_name='default', null=True, editable=False)),
                ('primary_key', models.BooleanField(default=False, verbose_name='primary key')),
                ('unique', models.BooleanField(default=False, verbose_name='unique')),
                ('unique_for_date', mutant.db.fields.python.PythonIdentifierField(max_length=255, null=True, verbose_name='unique for date', blank=True)),
                ('unique_for_month', mutant.db.fields.python.PythonIdentifierField(max_length=255, null=True, verbose_name='unique for month', blank=True)),
                ('unique_for_year', mutant.db.fields.python.PythonIdentifierField(max_length=255, null=True, verbose_name='unique for year', blank=True)),
            ],
            options={
                'verbose_name': 'field',
                'verbose_name_plural': 'fields',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='FieldDefinitionChoice',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('order', models.PositiveIntegerField(editable=False)),
                ('group', mutant.db.fields.translation.LazilyTranslatedField(null=True, verbose_name='group', blank=True)),
                ('value', picklefield.fields.PickledObjectField(verbose_name='value', editable=False)),
                ('label', mutant.db.fields.translation.LazilyTranslatedField(verbose_name='label')),
                ('field_def', models.ForeignKey(
                    to='mutant.FieldDefinition', on_delete=models.CASCADE, related_name='choices'
                )),
            ],
            options={
                'ordering': ['order'],
                'verbose_name': 'field definition choice',
                'verbose_name_plural': 'field definition choices',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='ModelDefinition',
            fields=[
                ('contenttype_ptr', models.OneToOneField(
                    to='contenttypes.ContentType', on_delete=models.CASCADE, parent_link=True, auto_created=True,
                    primary_key=True, serialize=False
                )),
                ('object_name', mutant.db.fields.python.PythonIdentifierField(max_length=255, verbose_name='object name')),
                ('db_table', models.CharField(max_length=63, null=True, verbose_name='database table', blank=True)),
                ('managed', models.BooleanField(default=False, verbose_name='managed')),
                ('verbose_name', mutant.db.fields.translation.LazilyTranslatedField(null=True, verbose_name='verbose name', blank=True)),
                ('verbose_name_plural', mutant.db.fields.translation.LazilyTranslatedField(null=True, verbose_name='verbose name plural', blank=True)),
            ],
            options={
                'verbose_name': 'model definition',
                'verbose_name_plural': 'model definitions',
            },
            bases=('contenttypes.contenttype',),
        ),
        migrations.CreateModel(
            name='OrderingFieldDefinition',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('order', models.PositiveIntegerField(editable=False)),
                ('lookup', models.CharField(max_length=255)),
                ('descending', models.BooleanField(default=False, verbose_name='descending')),
                ('model_def', models.ForeignKey(related_name='orderingfielddefinitions', on_delete=mutant.db.deletion.CASCADE_MARK_ORIGIN, to='mutant.ModelDefinition')),
            ],
            options={
                'ordering': ['order'],
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='UniqueTogetherDefinition',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('field_defs', models.ManyToManyField(related_name='unique_together_defs', to='mutant.FieldDefinition')),
                ('model_def', models.ForeignKey(related_name='uniquetogetherdefinitions', on_delete=mutant.db.deletion.CASCADE_MARK_ORIGIN, to='mutant.ModelDefinition')),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.AlterUniqueTogether(
            name='fielddefinitionchoice',
            unique_together=set([('field_def', 'order'), ('field_def', 'group', 'value')]),
        ),
        migrations.AddField(
            model_name='fielddefinition',
            name='content_type',
            field=mutant.db.fields.generic.FieldDefinitionTypeField(),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='fielddefinition',
            name='model_def',
            field=models.ForeignKey(related_name='fielddefinitions', on_delete=mutant.db.deletion.CASCADE_MARK_ORIGIN, to='mutant.ModelDefinition'),
            preserve_default=True,
        ),
        migrations.AlterUniqueTogether(
            name='fielddefinition',
            unique_together=set([('model_def', 'name')]),
        ),
        migrations.AddField(
            model_name='basedefinition',
            name='model_def',
            field=models.ForeignKey(related_name='basedefinitions', on_delete=mutant.db.deletion.CASCADE_MARK_ORIGIN, to='mutant.ModelDefinition'),
            preserve_default=True,
        ),
        migrations.AlterUniqueTogether(
            name='basedefinition',
            unique_together=set([('model_def', 'order')]),
        ),
    ]



from __future__ import unicode_literals

from django.core.cache import caches

from ...settings import STATE_CACHE_ALIAS


class CacheStateHandler(object):
    """
    State handlers that relies on cache to store and retrieve the current
    checksum of a definition."
    """

    def __init__(self):
        self.cache = caches[STATE_CACHE_ALIAS]

    def get_cache_key(self, definition_pk):
        return "mutant-%s" % definition_pk

    def get_checksum(self, definition_pk):
        cache_key = self.get_cache_key(definition_pk)
        return self.cache.get(cache_key)

    def set_checksum(self, definition_pk, checksum):
        cache_key = self.get_cache_key(definition_pk)
        return self.cache.set(cache_key, checksum)

    def clear_checksum(self, definition_pk):
        cache_key = self.get_cache_key(definition_pk)
        return self.cache.delete(cache_key)



from __future__ import unicode_literals

from django.db import models
from django.db.models.aggregates import Max


class OrderedModel(models.Model):
    order = models.PositiveIntegerField(editable=False)

    class Meta:
        abstract = True
        ordering = ['order']

    def get_ordering_queryset(self):
        return self.__class__._default_manager.all()

    def save(self, *args, **kwargs):
        if self.order is None:
            max_order = self.get_ordering_queryset().aggregate(
                Max('order')
            ).get('order__max')
            self.order = 0 if max_order is None else max_order + 1
        return super(OrderedModel, self).save(*args, **kwargs)



from __future__ import unicode_literals

from threading import RLock


class MemoryStateHandler(object):
    """
    State handler that relies on a lock and an in-memory map of definition
    pk and their associated checksums to maintain the current state of mutable
    models.
    """

    def __init__(self):
        self.lock = RLock()
        self.checksums = {}

    def flush(self):
        with self.lock:
            self.checksums.clear()

    def get_checksum(self, definition_pk):
        return self.checksums.get(definition_pk)

    def set_checksum(self, definition_pk, checksum):
        with self.lock:
            self.checksums[definition_pk] = checksum

    def clear_checksum(self, definition_pk):
        with self.lock:
            try:
                del self.checksums[definition_pk]
            except KeyError:
                pass



from __future__ import unicode_literals

from collections import defaultdict
from contextlib import contextmanager
from copy import deepcopy
from itertools import chain, groupby
from operator import itemgetter

from django.apps import AppConfig, apps
from django.db import connections, models, router
from django.utils import six
from django.utils.encoding import force_text
from django.utils.functional import lazy

from .compat import get_remote_field, get_remote_field_model


def allow_migrate(model):
    for db in connections:
        if router.allow_migrate(db, model):
            yield db


NOT_PROVIDED = object()


def popattr(obj, attr, default=NOT_PROVIDED):
    """
    Useful for retrieving an object attr and removing it if it's part of it's
    dict while allowing retrieving from subclass.
    i.e.
    class A:
        a = 'a'
    class B(A):
        b = 'b'
    >>> popattr(B, 'a', None)
    'a'
    >>> A.a
    'a'
    """
    val = getattr(obj, attr, default)
    try:
        delattr(obj, attr)
    except AttributeError:
        if default is NOT_PROVIDED:
            raise
    return val


def _string_format(string, *args, **kwargs):
    if args:
        return string % tuple(force_text(s) for s in args)
    elif kwargs:
        return string % dict((k, force_text(v)) for k, v in kwargs.items())
lazy_string_format = lazy(_string_format, six.text_type)


def get_db_table(app_label, model):
    return "mutant_%s_%s" % (app_label, model)


@contextmanager
def apps_lock():
    # The registry lock is not re-entrant so we must avoid acquiring it
    # during the initialization phase in order to prevent deadlocks.
    if apps.ready:
        with apps._lock:
            yield
    else:
        yield


def remove_from_app_cache(model_class, quiet=False):
    opts = model_class._meta
    apps = opts.apps
    app_label, model_name = opts.app_label, opts.model_name
    with apps_lock():
        try:
            model_class = apps.app_configs[app_label].models.pop(model_name)
        except KeyError:
            if not quiet:
                raise ValueError("%r is not cached" % model_class)
        apps.clear_cache()
        unreference_model(model_class)
    return model_class


def get_foward_fields(opts):
    return chain(
        opts.fields,
        opts.many_to_many
    )


def get_reverse_fields(opts):
    return opts._get_fields(forward=False, reverse=True, include_hidden=True)


def clear_opts_related_cache(model_class):
    opts = model_class._meta
    children = [
        related_object.related_model
        for related_object in opts.__dict__.get('related_objects', []) if related_object.parent_link
    ]
    opts._expire_cache()
    for child in children:
        clear_opts_related_cache(child)


def unreference_model(model):
    for field in get_foward_fields(model._meta):
        remote_field = get_remote_field(field)
        if field.model is model and remote_field:
            remote_field_model = get_remote_field_model(field)
            if isinstance(remote_field_model, models.base.ModelBase):
                clear_opts_related_cache(remote_field_model)
                rel_is_hidden = remote_field.is_hidden()
                # An accessor is added to related classes if they are not
                # hidden. However o2o fields *always* add an accessor
                # even if the relationship is hidden.
                o2o = isinstance(field, models.OneToOneField)
                if not rel_is_hidden or o2o:
                    try:
                        delattr(remote_field_model, get_remote_field(field).get_accessor_name())
                    except AttributeError:
                        # Hidden related names are not respected for o2o
                        # thus a tenant models with a o2o pointing to
                        # a non-tenant one would have a class for multiple
                        # tenant thus the attribute might be attempted
                        # to be deleted multiple times.
                        if not (o2o and rel_is_hidden):
                            raise


class Empty(object):
    pass


def _app_cache_deepcopy(obj):
    """
    An helper that correctly deepcopy model cache state
    """
    if isinstance(obj, defaultdict):
        return deepcopy(obj)
    elif isinstance(obj, dict):
        return type(obj)((_app_cache_deepcopy(key), _app_cache_deepcopy(val)) for key, val in obj.items())
    elif isinstance(obj, list):
        return list(_app_cache_deepcopy(val) for val in obj)
    elif isinstance(obj, AppConfig):
        app_conf = Empty()
        app_conf.__class__ = AppConfig
        app_conf.__dict__ = _app_cache_deepcopy(obj.__dict__)
        return app_conf
    return obj


@contextmanager
def app_cache_restorer():
    """
    A context manager that restore model cache state as it was before
    entering context.
    """
    state = _app_cache_deepcopy(apps.__dict__)
    try:
        yield state
    finally:
        with apps_lock():
            apps.__dict__ = state
            # Rebind the app registry models cache to
            # individual app config ones.
            for app_conf in apps.get_app_configs():
                app_conf.models = apps.all_models[app_conf.label]
            apps.clear_cache()


group_item_getter = itemgetter('group')


def choices_from_dict(choices):
    for grp, choices in groupby(choices, key=group_item_getter):
        if grp is None:
            for choice in choices:
                yield (choice['value'], choice['label'])
        else:
            yield (grp, tuple(
                (choice['value'], choice['label']) for choice in choices)
            )



from __future__ import unicode_literals

from django import forms
from django.contrib.contenttypes.models import ContentType
from django.utils.encoding import smart_text
from django.utils.functional import LazyObject

from .utils import choices_from_dict


class LazyFieldDefinitionQueryset(LazyObject):
    def __init__(self, queryset, models):
        super(LazyFieldDefinitionQueryset, self).__init__()
        self.__dict__.update(queryset=queryset, models=models)

    def _setup(self):
        queryset = self.__dict__.get('queryset')
        models = self.__dict__.get('models')
        self._wrapped = queryset.filter(
            pk__in=[ct.pk for ct in ContentType.objects.get_for_models(
                *models, for_concrete_models=False
            ).values()]
        )


class LazyFieldDefinitionGroupedChoices(LazyObject):
    def __init__(self, queryset, empty_label, label_from_instance):
        super(LazyFieldDefinitionGroupedChoices, self).__init__()
        self.__dict__.update(
            queryset=queryset, empty_label=empty_label,
            label_from_instance=label_from_instance
        )

    def _setup(self):
        queryset = self.__dict__.get('queryset')
        label_from_instance = self.__dict__.get('label_from_instance')
        empty_label = self.__dict__.get('empty_label')
        definition_choices = []
        for content_type in queryset:
            definition = content_type.model_class()
            category = definition.get_field_category()
            definition_choices.append({
                'group': smart_text(category) if category else None,
                'value': content_type.pk,
                'label': label_from_instance(content_type),
            })
        choices = list(
            choices_from_dict(
                sorted(definition_choices, key=lambda c: c['group'] or '')
            )
        )
        if empty_label is not None:
            choices.insert(0, ('', self.empty_label))
        self._wrapped = choices


class FieldDefinitionTypeField(forms.ModelChoiceField):
    def __init__(self, *args, **kwargs):
        self.field_definitions = kwargs.pop('field_definitions', [])
        self.group_by_category = kwargs.pop('group_by_category', False)
        super(FieldDefinitionTypeField, self).__init__(*args, **kwargs)

    def _get_field_definitions(self):
        return self._field_definitions

    def _set_field_definitions(self, definitions):
        for definition in definitions:
            from mutant.models import FieldDefinition
            if not issubclass(definition, FieldDefinition):
                raise TypeError(
                    "%r is not a subclass of `FieldDefinition`" % definition
                )
        self._field_definitions = definitions

    field_definitions = property(_get_field_definitions, _set_field_definitions)

    def _get_queryset(self):
        queryset = super(FieldDefinitionTypeField, self)._get_queryset()
        if self.field_definitions:
            return LazyFieldDefinitionQueryset(queryset, self.field_definitions)
        return queryset

    queryset = property(_get_queryset, forms.ModelChoiceField._set_queryset)

    def _get_choices(self):
        if self.group_by_category:
            return LazyFieldDefinitionGroupedChoices(
                self.queryset, self.empty_label, self.label_from_instance
            )
        return super(FieldDefinitionTypeField, self)._get_choices()

    choices = property(_get_choices, forms.ModelChoiceField._set_queryset)

    def label_from_instance(self, obj):
        return smart_text(obj.model_class().get_field_description())



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models

import mutant.db.fields.python


class Migration(migrations.Migration):

    dependencies = [
        ('text', '0002_update_field_defs_app_label'),
    ]

    operations = [
        migrations.CreateModel(
            name='FilePathFieldDefinition',
            fields=[
                ('charfielddefinition_ptr', models.OneToOneField(
                    to='text.CharFieldDefinition', on_delete=models.CASCADE, parent_link=True,
                    auto_created=True, primary_key=True, serialize=False
                )),
                ('path', mutant.db.fields.python.DirectoryPathField(help_text='The absolute filesystem path to a directory from which this field should get its choices.', max_length=100, verbose_name='path')),
                ('match', mutant.db.fields.python.RegExpStringField(help_text='A regular expression used to filter filenames.', max_length=100, null=True, verbose_name='match', blank=True)),
                ('recursive', models.BooleanField(default=False, help_text='Specifies whether all subdirectories of path should be included', verbose_name='recursive')),
            ],
            options={
                'db_table': 'mutant_filepathfielddefinition',
            },
            bases=('text.charfielddefinition',),
        ),
    ]



from __future__ import unicode_literals

import time

from django.utils.module_loading import import_string

from mutant import settings

from ..memory import MemoryStateHandler


class PubSubStateHandler(MemoryStateHandler):
    def __init__(self, engine=None):
        super(PubSubStateHandler, self).__init__()
        dotted_path, options = engine or settings.STATE_PUBSUB
        engine_cls = import_string(dotted_path)
        self.timestamps = {}
        self.engine = engine_cls(self.flush, self.receive, **options)
        self.engine.start()
        while not self.engine.ready:
            time.sleep(0.1)

    def receive(self, definition_pk, checksum, timestamp):
        # Do not alter current state if the change was published before our
        if self.timestamps.get(definition_pk, timestamp) > timestamp:
            return
        if checksum is None:
            super(PubSubStateHandler, self).clear_checksum(definition_pk)
        else:
            super(PubSubStateHandler, self).set_checksum(definition_pk, checksum)

    def set_checksum(self, definition_pk, checksum):
        timestamp = time.time()
        with self.lock:
            self.timestamps[definition_pk] = timestamp
            super(PubSubStateHandler, self).set_checksum(definition_pk, checksum)
        self.engine.publish(definition_pk, checksum, timestamp)

    def clear_checksum(self, definition_pk):
        timestamp = time.time()
        with self.lock:
            self.timestamps[definition_pk] = timestamp
            super(PubSubStateHandler, self).clear_checksum(definition_pk)
        self.engine.publish(definition_pk, None, timestamp)



from __future__ import unicode_literals

import django
from django.contrib.gis.db import models

GeoManager = models.Manager if django.VERSION >= (1, 9) else models.GeoManager


class GeoModel(models.Model):
    """
    A model to be used as a BaseDefinition on ModelDefinition instance with
    GeometryFieldDefinition instances.
    """
    objects = GeoManager()

    class Meta:
        abstract = True



from __future__ import unicode_literals

import warnings

import django
from django.db import models


class FilteredQuerysetManager(models.Manager):
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
        super(FilteredQuerysetManager, self).__init__()

    def get_queryset(self):
        qs = super(FilteredQuerysetManager, self).get_queryset()
        return qs.filter(*self.args, **self.kwargs)

    if django.VERSION < (1, 8):
        def get_query_set(self):
            warnings.warn(
                "`FilteredQuerysetManager.get_query_set` is deprecated, "
                "use `get_queryset` instead",
                DeprecationWarning if django.VERSION >= (1, 7)
                else PendingDeprecationWarning,
                stacklevel=2
            )
            return FilteredQuerysetManager.get_queryset(self)



from __future__ import unicode_literals

from django.conf import settings
from django.core.cache import DEFAULT_CACHE_ALIAS

STATE_HANDLER = getattr(
    settings, 'MUTANT_STATE_HANDLER',
    'mutant.state.handlers.memory.MemoryStateHandler'
)

STATE_CACHE_ALIAS = getattr(
    settings, 'MUTANT_STATE_CACHE_ALIAS', DEFAULT_CACHE_ALIAS
)

STATE_PUBSUB = getattr(
    settings, 'MUTANT_STATE_PUBSUB', (
        'mutant.state.handlers.pubsub.engines.Redis', {}
    )
)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import functools

from django.db import migrations


def _update_field_def_cts_app_label(from_app_label, to_app_label, apps, schema_editor):
    cts = apps.get_model('contenttypes', 'contenttype').objects
    cts.filter(
        app_label=from_app_label, model='filepathfielddefinition'
    ).update(app_label=to_app_label)


class Migration(migrations.Migration):

    dependencies = [
        ('file', '0001_initial'),
    ]

    operations = [
        migrations.RunPython(
            functools.partial(_update_field_def_cts_app_label, 'mutant', 'file'),
            functools.partial(_update_field_def_cts_app_label, 'file', 'mutant'),
        ),
        migrations.AlterModelTable(
            name='filepathfielddefinition',
            table=None,
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('mutant', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='DecimalFieldDefinition',
            fields=[
                ('fielddefinition_ptr', models.OneToOneField(
                    to='mutant.FieldDefinition', on_delete=models.CASCADE, parent_link=True, auto_created=True,
                    primary_key=True, serialize=False
                )),
                ('max_digits', models.PositiveSmallIntegerField(help_text='The maximum number of digits allowed in the number. Note that this number must be greater than or equal to ``decimal_places``, if it exists.', verbose_name='max digits')),
                ('decimal_places', models.PositiveSmallIntegerField(help_text='The number of decimal places to store with the number.', verbose_name='decimal_places')),
            ],
            options={
                'db_table': 'mutant_decimalfielddefinition',
            },
            bases=('mutant.fielddefinition',),
        ),
        migrations.CreateModel(
            name='BigIntegerFieldDefinition',
            fields=[
            ],
            options={
                'proxy': True,
            },
            bases=('mutant.fielddefinition',),
        ),
        migrations.CreateModel(
            name='FloatFieldDefinition',
            fields=[
            ],
            options={
                'proxy': True,
            },
            bases=('mutant.fielddefinition',),
        ),
        migrations.CreateModel(
            name='IntegerFieldDefinition',
            fields=[
            ],
            options={
                'proxy': True,
            },
            bases=('mutant.fielddefinition',),
        ),
        migrations.CreateModel(
            name='PositiveIntegerFieldDefinition',
            fields=[
            ],
            options={
                'proxy': True,
            },
            bases=('mutant.fielddefinition',),
        ),
        migrations.CreateModel(
            name='PositiveSmallIntegerFieldDefinition',
            fields=[
            ],
            options={
                'proxy': True,
            },
            bases=('mutant.fielddefinition',),
        ),
        migrations.CreateModel(
            name='SmallIntegerFieldDefinition',
            fields=[
            ],
            options={
                'proxy': True,
            },
            bases=('mutant.fielddefinition',),
        ),
    ]



from __future__ import unicode_literals

from django.db.models.fields import TextField
from django.utils.encoding import smart_text
from django.utils.functional import Promise
from django.utils.translation import ugettext_lazy as _


def _is_gettext_promise(value):
    return isinstance(value, Promise) and (value._delegate_bytes or value._delegate_text)


class LazilyTranslatedField(TextField):
    def to_python(self, value):
        if value is None or _is_gettext_promise(value):
            return value
        return _(smart_text(value))

    def get_prep_value(self, value):
        if value is None:
            return value
        elif _is_gettext_promise(value):
            value = smart_text(value._proxy____args[0])
        return smart_text(value)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('text', '0002_update_field_defs_app_label'),
    ]

    operations = [
        migrations.CreateModel(
            name='GenericIPAddressFieldDefinition',
            fields=[
                ('charfielddefinition_ptr', models.OneToOneField(
                    to='text.CharFieldDefinition', on_delete=models.CASCADE, parent_link=True, auto_created=True,
                    primary_key=True, serialize=False
                )),
                ('protocol', models.CharField(default='both', max_length=4, verbose_name='protocol', choices=[('both', 'both'), ('IPv4', 'IPv4'), ('IPv6', 'IPv6')])),
                ('unpack_ipv4', models.BooleanField(default=False, verbose_name='unpack ipv4')),
            ],
            options={
                'db_table': 'mutant_genericipaddressfielddefinition',
            },
            bases=('text.charfielddefinition',),
        ),
        migrations.CreateModel(
            name='EmailFieldDefinition',
            fields=[
            ],
            options={
                'proxy': True,
            },
            bases=('text.charfielddefinition',),
        ),
        migrations.CreateModel(
            name='IPAddressFieldDefinition',
            fields=[
            ],
            options={
                'proxy': True,
            },
            bases=('text.charfielddefinition',),
        ),
        migrations.CreateModel(
            name='SlugFieldDefinition',
            fields=[
            ],
            options={
                'proxy': True,
            },
            bases=('text.charfielddefinition',),
        ),
        migrations.CreateModel(
            name='URLFieldDefinition',
            fields=[
            ],
            options={
                'proxy': True,
            },
            bases=('text.charfielddefinition',),
        ),
    ]



from __future__ import unicode_literals

from django.db import models
from polymodels.fields import PolymorphicTypeField

from ... import forms


class FieldDefinitionTypeField(PolymorphicTypeField):
    def __init__(self, on_delete=models.CASCADE, *args, **kwargs):
        super(FieldDefinitionTypeField, self).__init__(
            'mutant.FieldDefinition', on_delete=on_delete, *args, **kwargs
        )

    def deconstruct(self):
        name, path, args, kwargs = super(FieldDefinitionTypeField, self).deconstruct()
        kwargs.pop('polymorphic_type')
        if kwargs.get('on_delete') == models.CASCADE:
            kwargs.pop('on_delete')
        return name, path, args, kwargs

    def formfield(self, **kwargs):
        defaults = {'form_class': forms.FieldDefinitionTypeField}
        defaults.update(kwargs)
        return super(FieldDefinitionTypeField, self).formfield(**kwargs)



from __future__ import unicode_literals

from functools import wraps

from django.contrib.contenttypes.models import ContentType
from django.db import connections, models, transaction
from django.db.models.fields import FieldDoesNotExist

from ..compat import get_remote_field
from ..state import handler as state_handler
from ..utils import allow_migrate, popattr, remove_from_app_cache


def perform_ddl(action, model, *args, **kwargs):
    if model._meta.managed:
        return

    for alias in allow_migrate(model):
        connection = connections[alias]
        with transaction.atomic(alias), connection.schema_editor() as editor:
            getattr(editor, action)(model, *args, **kwargs)


def nonraw_instance(receiver):
    """
    A signal receiver decorator that fetch the complete instance from db when
    it's passed as raw
    """
    @wraps(receiver)
    def wrapper(sender, instance, raw, using, **kwargs):
        if raw:
            instance = sender._default_manager.using(using).get(pk=instance.pk)
        return receiver(sender=sender, raw=raw, instance=instance, using=using,
                        **kwargs)
    return wrapper


@nonraw_instance
def model_definition_post_save(sender, instance, created, **kwargs):
    model_class = instance.model_class(force_create=True)
    opts = model_class._meta
    db_table = opts.db_table
    if created:
        primary_key = opts.pk
        fields = [(field.get_attname_column()[1], field) for field in opts.fields
                  if field is not primary_key]
        try:
            extra_fields = getattr(instance._state, '_create_extra_fields')
        except AttributeError:
            pass
        else:
            for column, field in extra_fields:
                remote_field = get_remote_field(field)
                if field.primary_key:
                    assert isinstance(primary_key, models.AutoField)
                    primary_key = field
                elif (remote_field and remote_field.parent_link and
                      isinstance(primary_key, models.AutoField)):
                    field.primary_key = True
                    primary_key = field
                else:
                    fields.append((column, field))
            delattr(instance._state, '_create_extra_fields')
        fields.insert(0, (primary_key.get_attname_column()[1], primary_key))
        try:
            delayed_save = getattr(instance._state, '_create_delayed_save')
        except AttributeError:
            pass
        else:
            for obj in delayed_save:
                obj.model_def = instance
                obj.save(force_insert=True, force_create_model_class=False)
            delattr(instance._state, '_create_delayed_save')
        model_class = instance.model_class(force_create=True)
        perform_ddl('create_model', model_class)
    else:
        old_model_class = instance._model_class
        if old_model_class:
            old_db_table = old_model_class._meta.db_table
            if db_table != old_db_table:
                perform_ddl('alter_db_table', model_class, old_db_table, db_table)
            ContentType.objects.clear_cache()
    instance._model_class = model_class.model


def model_definition_pre_delete(sender, instance, **kwargs):
    model_class = instance.model_class()
    instance._state._deletion = (
        model_class,
        instance.pk,
    )


def model_definition_post_delete(sender, instance, **kwargs):
    model_class, pk = popattr(instance._state, '_deletion')
    perform_ddl('delete_model', model_class)
    remove_from_app_cache(model_class)
    model_class.mark_as_obsolete()
    state_handler.clear_checksum(pk)
    ContentType.objects.clear_cache()
    del instance._model_class


def base_definition_post_save(sender, instance, created, raw, **kwargs):
    declared_fields = instance.get_declared_fields()
    if declared_fields:
        model_class = instance.model_def.model_class().render_state()
        opts = model_class._meta
        if created:
            add_columns = popattr(instance._state, '_add_columns', True)
            if add_columns:
                auto_pk = isinstance(opts.pk, models.AutoField)
                for field in declared_fields:
                    field.model = model_class
                    remote_field = get_remote_field(field)
                    if auto_pk and remote_field and remote_field.parent_link:
                        auto_pk = False
                        field.primary_key = True
                        perform_ddl('alter_field', model_class, opts.pk, field, strict=True)
                    else:
                        perform_ddl('add_field', model_class, field)
        else:
            for field in declared_fields:
                try:
                    old_field = opts.get_field(field.name)
                except FieldDoesNotExist:
                    perform_ddl('add_field', model_class, field)
                else:
                    perform_ddl('alter_field', model_class, old_field, field, strict=True)


def base_definition_pre_delete(sender, instance, **kwargs):
    """
    This is used to pass data required for deletion to the post_delete
    signal that is no more available thereafter.
    """
    # see CASCADE_MARK_ORIGIN's docstring
    cascade_deletion_origin = popattr(
        instance._state, '_cascade_deletion_origin', None
    )
    if cascade_deletion_origin == 'model_def':
        return
    if (instance.base and issubclass(instance.base, models.Model) and
            instance.base._meta.abstract):
        instance._state._deletion = instance.model_def.model_class().render_state()


def base_definition_post_delete(sender, instance, **kwargs):
    """
    Make sure to delete fields inherited from an abstract model base.
    """
    if hasattr(instance._state, '_deletion'):
        # Make sure to flatten abstract bases since Django
        # migrations can't deal with them.
        model = popattr(instance._state, '_deletion')
        for field in instance.base._meta.fields:
            perform_ddl('remove_field', model, field)


def unique_together_field_defs_changed(instance, action, model, **kwargs):
    model_class = instance.model_def.model_class()
    if action.startswith('post_'):
        perform_ddl(
            'alter_unique_together',
            model_class,
            model_class._meta.unique_together,
            instance.model_def.get_state().options.get('unique_together', [])
        )
        model_class.mark_as_obsolete()


def raw_field_definition_proxy_post_save(sender, instance, raw, **kwargs):
    """
    When proxy field definitions are loaded from a fixture they're not
    passing through the `field_definition_post_save` signal. Make sure they
    are.
    """
    if raw:
        model_class = instance.content_type.model_class()
        opts = model_class._meta
        if opts.proxy and opts.concrete_model is sender:
            field_definition_post_save(
                sender=model_class, instance=instance.type_cast(), raw=raw,
                **kwargs
            )


@nonraw_instance
def field_definition_post_save(sender, instance, created, raw, **kwargs):
    """
    This signal is connected by all FieldDefinition subclasses
    see comment in FieldDefinitionBase for more details
    """
    model_class = instance.model_def.model_class().render_state()
    field = instance.construct_for_migrate()
    field.model = model_class
    if created:
        if hasattr(instance._state, '_creation_default_value'):
            field.default = instance._state._creation_default_value
            delattr(instance._state, '_creation_default_value')
        add_column = popattr(instance._state, '_add_column', True)
        if add_column:
            perform_ddl('add_field', model_class, field)
            # If the field definition is raw we must re-create the model class
            # since ModelDefinitionAttribute.save won't be called
            if raw:
                instance.model_def.model_class().mark_as_obsolete()
    else:
        old_field = instance._state._pre_save_field
        delattr(instance._state, '_pre_save_field')
        perform_ddl('alter_field', model_class, old_field, field, strict=True)

FIELD_DEFINITION_POST_SAVE_UID = "mutant.management.%s_post_save"


def field_definition_pre_delete(sender, instance, **kwargs):
    # see CASCADE_MARK_ORIGIN's docstring
    cascade_deletion_origin = popattr(
        instance._state, '_cascade_deletion_origin', None
    )
    if cascade_deletion_origin == 'model_def':
        return
    model_class = instance.model_def.model_class()
    opts = model_class._meta
    field = opts.get_field(instance.name)
    instance._state._deletion = (model_class, field)


def field_definition_post_delete(sender, instance, **kwargs):
    if hasattr(instance._state, '_deletion'):
        model, field = popattr(instance._state, '_deletion')
        if field.primary_key:
            primary_key = models.AutoField(name='id', primary_key=True)
            perform_ddl('alter_field', model, field, primary_key, strict=True)
        else:
            perform_ddl('remove_field', model, field)



from __future__ import unicode_literals

from operator import attrgetter

import django

get_remote_field = attrgetter('remote_field' if django.VERSION >= (1, 9) else 'rel')

if django.VERSION >= (1, 9):
    def get_remote_field_model(field):
        model = getattr(field, 'model', None)
        if model:
            return field.remote_field.model
        else:
            return field.related_model

    def get_opts_label(opts):
        return opts.label
else:
    def get_remote_field_model(field):
        return getattr(getattr(field, 'rel', None), 'to', None)

    def get_opts_label(opts):
        return "%s.%s" % (opts.app_label, opts.object_name)



#!/usr/bin/env python
import os
from setuptools import find_packages, setup

from mutant import __version__

MODULE_PATH = os.path.abspath(os.path.dirname(__file__))

setup(
    name='django-mutant',
    version=__version__,
    description='Dynamic model definition and alteration (evolving schemas)',
    long_description=open(os.path.join(MODULE_PATH, 'README.rst')).read(),
    url='https://github.com/charettes/django-mutant',
    author='Simon Charette',
    author_email='charette.s@gmail.com',
    install_requires=[
        'django>=1.8',
        'django-picklefield>=0.3.2',
        'django-polymodels>=1.4.6a1',
    ],
    packages=find_packages(exclude=['tests', 'tests.*']),
    package_data={
        '': ['locale/*/LC_MESSAGES/*'],
    },
    license='MIT License',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 1.8',
        'Framework :: Django :: 1.9',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ]
)



from __future__ import unicode_literals

from django.apps import apps


class HandlerProxy(object):
    def __getattribute__(self, name):
        if not apps.apps_ready:
            raise AttributeError
        handler = apps.get_app_config('mutant').state_handler
        return getattr(handler, name)



from __future__ import unicode_literals

from django.apps import AppConfig
from django.db import models
from django.utils.module_loading import import_string

from . import settings


class MutantConfig(AppConfig):
    name = 'mutant'

    def ready(self):
        self.state_handler = import_string(settings.STATE_HANDLER)()

        from . import management

        ModelDefinition = self.get_model('ModelDefinition')
        models.signals.post_save.connect(
            management.model_definition_post_save,
            sender=ModelDefinition,
            dispatch_uid='mutant.management.model_definition_post_save',
        )
        models.signals.pre_delete.connect(
            management.model_definition_pre_delete,
            sender=ModelDefinition,
            dispatch_uid='mutant.management.model_definition_pre_delete',
        )
        models.signals.post_delete.connect(
            management.model_definition_post_delete,
            sender=ModelDefinition,
            dispatch_uid='mutant.management.model_definition_post_delete',
        )

        BaseDefinition = self.get_model('BaseDefinition')
        models.signals.post_save.connect(
            management.base_definition_post_save,
            sender=BaseDefinition,
            dispatch_uid='mutant.management.base_definition_post_save',
        )
        models.signals.pre_delete.connect(
            management.base_definition_pre_delete,
            sender=BaseDefinition,
            dispatch_uid='mutant.management.base_definition_pre_delete',
        )
        models.signals.post_delete.connect(
            management.base_definition_post_delete,
            sender=BaseDefinition,
            dispatch_uid='mutant.management.base_definition_post_delete',
        )

        UniqueTogetherDefinition = self.get_model('UniqueTogetherDefinition')
        models.signals.m2m_changed.connect(
            management.unique_together_field_defs_changed,
            sender=UniqueTogetherDefinition.field_defs.through,
            dispatch_uid='mutant.management.unique_together_field_defs_changed',
        )

        FieldDefinition = self.get_model('FieldDefinition')
        models.signals.post_save.connect(
            management.raw_field_definition_proxy_post_save,
            sender=FieldDefinition,
            dispatch_uid='mutant.management.raw_field_definition_proxy_post_save',
        )
        models.signals.pre_delete.connect(
            management.field_definition_pre_delete,
            sender=FieldDefinition,
            dispatch_uid='mutant.management.field_definition_pre_delete',
        )
        models.signals.post_delete.connect(
            management.field_definition_post_delete,
            sender=FieldDefinition,
            dispatch_uid='mutant.management.field_definition_post_delete',
        )



from __future__ import unicode_literals

import pickle
from hashlib import md5

from django.apps import apps
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError
from django.db import models
from django.db.migrations.state import ModelState
from django.db.models.constants import LOOKUP_SEP
from django.db.models.fields import FieldDoesNotExist
from django.utils.encoding import python_2_unicode_compatible
from django.utils.translation import ugettext_lazy as _
from picklefield.fields import PickledObjectField

from ... import logger
from ...compat import get_opts_label, get_remote_field_model
from ...db.deletion import CASCADE_MARK_ORIGIN
from ...db.fields import LazilyTranslatedField, PythonIdentifierField
from ...db.models import MutableModel
from ...signals import mutable_class_prepared
from ...state import handler as state_handler
from ...utils import get_db_table, get_foward_fields, remove_from_app_cache
from ..ordered import OrderedModel
from .managers import ModelDefinitionManager


def _model_class_from_pk(definition_cls, definition_pk):
    """
    Helper used to unpickle MutableModel model class from their definition
    pk.
    """
    try:
        return definition_cls.objects.get(pk=definition_pk).model_class()
    except definition_cls.DoesNotExist:
        pass


class MutableModelProxy(object):
    __slots__ = ['model', 'refreshing', '__weakref__']

    proxied_methods = [
        '__setattr__', '__delattr__', '__hash__', '__str__'
    ]

    @classmethod
    def method_factory(cls, name):
        def method(self, *args, **kwargs):
            model = self.__get__()
            return getattr(model.__class__, name)(model, *args, **kwargs)
        method.__name__ = str(name)
        return method

    @classmethod
    def factory(cls, base):
        attrs = dict(
            (name, cls.method_factory(name)) for name in cls.proxied_methods
        )
        name = str("%s(%s)" % (cls.__name__, base.__name__))
        return type(name, (cls,), attrs)

    def __new__(cls, model, *args, **kwargs):
        base = model.__class__
        try:
            cache = cls._proxy_class_cache
        except AttributeError:
            cls._proxy_class_cache = cache = {}
        try:
            proxy_class = cache[base]
        except KeyError:
            cache[base] = proxy_class = cls.factory(base)
        return super(MutableModelProxy, cls).__new__(proxy_class)

    def __init__(self, model):
        assert issubclass(model, MutableModel)
        supset = super(MutableModelProxy, self).__setattr__
        supset('model', model)
        supset('refreshing', False)

    def __get__(self, instance=None, owner=None):
        model = self.model
        if not self.refreshing and model.is_obsolete():
            supset = super(MutableModelProxy, self).__setattr__
            try:
                supset('refreshing', True)
                try:
                    definition = model.definition()
                finally:
                    supset('refreshing', False)
            except ModelDefinition.DoesNotExist:
                raise AttributeError(
                    'The definition of %s.%s has been deleted.' % (
                        model._meta.app_label,
                        model._meta.object_name,
                    )
                )
            else:
                proxy = definition.model_class()
                assert isinstance(proxy, MutableModelProxy)
                model = proxy.model
                supset('model', model)
        return model

    def __getattribute__(self, name):
        if name in ('model', 'refreshing', '__get__', '__eq__', '__ne__', '__reduce_ex__'):
            return super(MutableModelProxy, self).__getattribute__(name)
        model = super(MutableModelProxy, self).__getattribute__('__get__')()
        return getattr(model, name)

    def __call__(self, *args, **kwargs):
        model = self.__get__()
        return model(*args, **kwargs)

    def __eq__(self, other):
        model = self.__get__()
        if isinstance(other, MutableModelProxy):
            other = other.__get__()
        if type(model) == type(other):
            return model == other
        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __instancecheck__(self, instance):
        model = self.__get__()
        return isinstance(instance, model)

    def __reduce_ex__(self, protocol):
        model = self.__get__()
        return (_model_class_from_pk, model._definition)


@python_2_unicode_compatible
class ModelDefinition(ContentType):
    object_name = PythonIdentifierField(_('object name'))
    db_table = models.CharField(
        _('database table'), max_length=63, blank=True, null=True
    )
    managed = models.BooleanField(_('managed'), default=False)
    verbose_name = LazilyTranslatedField(
        _('verbose name'), blank=True, null=True
    )
    verbose_name_plural = LazilyTranslatedField(
        _('verbose name plural'), blank=True, null=True
    )

    objects = ModelDefinitionManager()

    class Meta:
        app_label = 'mutant'
        verbose_name = _('model definition')
        verbose_name_plural = _('model definitions')

    def __str__(self):
        return "%s.%s" % (self.app_label, self.object_name)

    def natural_key(self):
        return (self.app_label, self.model)
    natural_key.dependencies = ['contenttypes.contenttype']

    def __init__(self, *args, **kwargs):
        # Attach unsaved related objects
        bases = kwargs.pop('bases', ())
        fields = kwargs.pop('fields', ())
        extra_fields = []
        delayed_save = []
        for base in bases:
            assert base.pk is None, 'Cannot associate already existing BaseDefinition'
            extra_fields.extend(
                [(f.get_attname_column()[1], f) for f in base.get_declared_fields(fields)]
            )
            base._state._add_columns = False
            delayed_save.append(base)
        for field in fields:
            assert field.pk is None, 'Cannot associate already existing FieldDefinition'
            field_instance = field.construct_for_migrate()
            extra_fields.append(
                (field_instance.get_attname_column()[1], field_instance)
            )
            field._state._add_column = False
            delayed_save.append(field)
        super(ModelDefinition, self).__init__(*args, **kwargs)
        # Add those fields to the instance state to be retrieved later
        self._state._create_extra_fields = extra_fields
        self._state._create_delayed_save = delayed_save
        if self.pk:
            self._model_class = self.model_class().model

    def get_model_bases(self):
        """Build a tuple of bases for the constructed definition"""
        bases = []
        has_mutable_base = False
        for base_def in self.basedefinitions.all():
            base = base_def.construct()
            if issubclass(base, MutableModel):
                has_mutable_base = True
                base._dependencies.add((self.__class__, self.pk))
            bases.append(base)
        if not has_mutable_base:
            bases.append(MutableModel)
        return tuple(bases)

    def get_model_opts(self):
        opts = {'app_label': self.app_label, 'managed': self.managed}
        # Database table
        db_table = self.db_table
        if db_table is None:
            db_table = get_db_table(*self.natural_key())
        opts['db_table'] = db_table
        # Verbose names
        if self.verbose_name is not None:
            opts['verbose_name'] = self.verbose_name
        if self.verbose_name_plural is not None:
            opts['verbose_name_plural'] = self.verbose_name_plural
        # Unique together
        unique_together = [
            ut for ut in (
                ut_def.construct() for ut_def in self.uniquetogetherdefinitions.all()
            ) if ut
        ]
        if unique_together:
            opts['unique_together'] = unique_together
        # Ordering
        ordering = tuple(
            ord_field_def.construct()
            for ord_field_def in self.orderingfielddefinitions.all()
        )
        if ordering:
            # Make sure not to add ordering if it's empty since it would
            # prevent the model from inheriting its possible base ordering.
            # Kinda related to django #17429
            opts['ordering'] = ordering
        return opts

    def get_model_attrs(self):
        try:
            app = apps.get_app_config(self.app_label)
        except LookupError:
            __module__ = str("mutant.apps.%s.models" % self.app_label)
        else:
            __module__ = str("%s.models" % app.module.__name__)
        attrs = {
            '__module__': __module__,
            '_definition': (self.__class__, self.pk),
            '_dependencies': set(),
            '_is_obsolete': False,
        }
        return attrs

    def get_state(self):
        fields = [
            (field_def.name, field_def.construct()) for field_def in self.fielddefinitions.select_subclasses()
        ]
        options = self.get_model_opts()
        bases = self.get_model_bases()
        return ModelState(self.app_label, self.object_name, fields=fields, options=options, bases=bases)

    def construct(self, force_create=False, existing_model_class=None):
        state = self.get_state()
        attrs = self.get_model_attrs()

        identifier = (
            self.pk, self.object_name, state.options, dict(
                (name, field.deconstruct()) for name, field in state.fields
            ), [
                MutableModelProxy(base).checksum()
                if base is not MutableModel and issubclass(base, MutableModel) else base
                for base in state.bases
            ]
        )
        checksum = md5(pickle.dumps(identifier)).hexdigest()
        state_handler.set_checksum(self.pk, checksum)

        if existing_model_class:
            if not force_create and existing_model_class._checksum == checksum:
                existing_model_class._is_obsolete = False
                return existing_model_class
            remove_from_app_cache(existing_model_class)
            existing_model_class.mark_as_obsolete()

        try:
            model_class = state.render(apps)
        except RuntimeError:
            # Account for race conditions between the removal from the apps
            # and the rendering of the new state.
            return apps.get_model(state.app_label, state.name)
        model_class._checksum = checksum
        for attr, value in attrs.items():
            setattr(model_class, attr, value)

        mutable_class_prepared.send(
            sender=model_class, definition=self,
            existing_model_class=existing_model_class
        )
        logger.debug("Created model class %s.", model_class)

        return model_class

    def model_class(self, force_create=False):
        model_class = super(ModelDefinition, self).model_class()
        if force_create or model_class is None or model_class.is_obsolete():
            model_class = self.construct(force_create, model_class)
        return MutableModelProxy(model_class)

    @property
    def model_ct(self):
        try:
            content_type = self._contenttype_ptr_cache
        except AttributeError:
            content_type = ContentType.objects.get_for_id(
                self.contenttype_ptr_id
            )
            self._contenttype_ptr_cache = content_type
        return content_type

    def save(self, *args, **kwargs):
        self.model = self.object_name.lower()
        model_class = getattr(self, '_model_class', None)
        if model_class:
            remove_from_app_cache(model_class)
        return super(ModelDefinition, self).save(*args, **kwargs)


class ModelDefinitionAttribute(models.Model):
    """
    A mixin used to make sure models that alter the state of a defined model
    clear the cached version
    """
    model_def = models.ForeignKey(
        ModelDefinition, related_name="%(class)ss",
        on_delete=CASCADE_MARK_ORIGIN
    )

    class Meta:
        abstract = True

    def save(self, *args, **kwargs):
        force_create = kwargs.pop('force_create_model_class', True)
        save = super(ModelDefinitionAttribute, self).save(*args, **kwargs)
        self.model_def.model_class(force_create=force_create)
        return save

    def delete(self, *args, **kwargs):
        force_create = kwargs.pop('force_create_model_class', True)
        delete = super(ModelDefinitionAttribute, self).delete(*args, **kwargs)
        self.model_def.model_class(force_create=force_create)
        return delete


class OrderedModelDefinitionAttribute(OrderedModel, ModelDefinitionAttribute):
    class Meta:
        abstract = True

    def get_ordering_queryset(self):
        qs = super(OrderedModelDefinitionAttribute, self).get_ordering_queryset()
        return qs.filter(model_def_id=self.model_def_id)


class BaseDefinition(OrderedModelDefinitionAttribute):
    """
    Model used to represent bases of a ModelDefinition
    """
    base = PickledObjectField(_('base'))

    class Meta:
        app_label = 'mutant'
        ordering = ['order']
        unique_together = (('model_def', 'order'),)

    def clean(self):
        try:
            if issubclass(self.base, models.Model):
                if self.base._meta.proxy:
                    raise ValidationError(_("Base can't be a proxy model."))
                elif (issubclass(self.base, MutableModel) and
                      self.base.definition() == self.model_def):
                    raise ValidationError(_("A model definition can't be a base of itself."))
        except TypeError:
            raise ValidationError(_('Base must be a class.'))
        return super(BaseDefinition, self).clean()

    def construct(self):
        if isinstance(self.base, MutableModelProxy):
            return self.base.__get__()
        return self.base

    def get_declared_fields(self, field_defs=None):
        if field_defs is None:
            field_defs = self.model_def.fielddefinitions.select_subclasses()
        existing_fields = [field.construct() for field in field_defs]
        fields = []
        if issubclass(self.base, models.Model):
            opts = self.base._meta
            if opts.abstract:
                # Add fields inherited from base's abstract parent and
                # local fields.
                for field in get_foward_fields(opts):
                    if field.model is self.base or field.model._meta.abstract:
                        clone = field.clone()
                        clone.set_attributes_from_name(field.name)
                        fields.append(clone)
            elif not opts.proxy and not any(
                    isinstance(field, models.OneToOneField) and
                    field.rel.parent_link and
                    get_remote_field_model(field) == get_opts_label(opts)
                    for field in existing_fields):
                # This is a concrete model base, we must declare a o2o
                attr_name = '%s_ptr' % opts.model_name
                parent_link = models.OneToOneField(
                    self.base, on_delete=models.CASCADE,
                    name=attr_name, auto_created=True, parent_link=True
                )
                parent_link.set_attributes_from_name(attr_name)
                fields.append(parent_link)
        return tuple(fields)


class OrderingFieldDefinition(OrderedModelDefinitionAttribute):
    lookup = models.CharField(max_length=255)
    descending = models.BooleanField(_('descending'), default=False)

    class Meta:
        app_label = 'mutant'
        ordering = ['order']
        # TODO: Should be unique both it bugs order swapping
        # unique_together = (('model_def', 'order'),)

    def clean(self):
        """
        Make sure the lookup makes sense
        """
        if self.lookup == '?':  # Randomly sort
            return
        else:
            lookups = self.lookup.split(LOOKUP_SEP)
            opts = self.model_def.model_class()._meta
            valid = True
            while len(lookups):
                lookup = lookups.pop(0)
                try:
                    field = opts.get_field(lookup)
                except FieldDoesNotExist:
                    valid = False
                else:
                    if isinstance(field, models.ForeignKey):
                        opts = get_remote_field_model(field)._meta
                    elif len(lookups):  # Cannot go any deeper
                        valid = False
                finally:
                    if not valid:
                        msg = _("This field doesn't exist")
                        raise ValidationError({'lookup': [msg]})

    def construct(self):
        return ("-%s" % self.lookup) if self.descending else self.lookup


@python_2_unicode_compatible
class UniqueTogetherDefinition(ModelDefinitionAttribute):
    field_defs = models.ManyToManyField(
        'FieldDefinition', related_name='unique_together_defs'
    )

    class Meta:
        app_label = 'mutant'

    def __str__(self):
        if self.pk:
            names = ', '.join(self.construct())
            return _("Unique together of (%s)") % names
        return ''

    def clean(self):
        for field_def in self.field_defs.select_related('model_def'):
            if field_def.model_def != self.model_def:
                msg = _('All fields must be of the same model')
                raise ValidationError({'field_defs': [msg]})

    def construct(self):
        return tuple(self.field_defs.names())



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('mutant', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='CharFieldDefinition',
            fields=[
                ('fielddefinition_ptr', models.OneToOneField(
                    to='mutant.FieldDefinition', on_delete=models.CASCADE, parent_link=True, auto_created=True,
                    primary_key=True, serialize=False
                )),
                ('max_length', models.PositiveSmallIntegerField(null=True, verbose_name='max length', blank=True)),
            ],
            options={
                'db_table': 'mutant_charfielddefinition',
            },
            bases=('mutant.fielddefinition',),
        ),
        migrations.CreateModel(
            name='TextFieldDefinition',
            fields=[
            ],
            options={
                'proxy': True,
            },
            bases=('text.charfielddefinition',),
        ),
    ]



from __future__ import unicode_literals

from django.apps import AppConfig

from ...signals import mutable_class_prepared


class RelatedConfig(AppConfig):
    name = 'mutant.contrib.related'

    def ready(self):
        from . import management
        mutable_class_prepared.connect(management.mutable_model_prepared)
